@import url('https://fonts.googleapis.com/css2?family=Oswald&family=Roboto:ital,wght@0,400;0,700;1,400;1,700&display=swap');
:root {
    font-size: 107%;
}

.sidebar-brand-text {
    display: none;
}

h1 {
    /*font-family: "Oswald", sans-serif;*/
}
h2 {
    font-size: 1.8em;
}